import qiime2.plugin
from qiime2.plugin import (SemanticType, Str, Int, Float, Choices, List,
                           MetadataColumn, Categorical, Numeric, Metadata, Bool, Visualization)
from q2_types.feature_table import (
    FeatureTable, Frequency)
from q2_types.feature_data import (FeatureData, Differential, Sequence, Taxonomy, )

from . import __version__
from .methods import single_enrichment, enrichment, embed_seqs, bg_term_enrichment, diff_abundance, enrich_pipeline, trim_primers
from .visualizations import draw_wordcloud_vis, plot_enrichment, heatmap, venn


_citation = ('manuscript in preparation')

_short_description = "A Qiime2 plugin for the dbBact bacterial knowledge-base (dbbact.org)"


# options for the diff_abundance plugin
_statistical_tests = ['meandiff', 'stdmeandiff']
_transform_functions = ['rankdata', 'log2data', 'normdata', 'binarydata', 'none']
_fdr_method = ['dsfdr', 'bhfdr', 'byfdr', 'filterBH']


plugin = qiime2.plugin.Plugin(
    name='dbbact',
    version=__version__,
    website='https://github.com/amnona/q2_dbbact',
    package='q2_dbbact',
    short_description='dbBact plugin for Qiime2',
    description=('A Qiime2 plugin for the dbBact bacterial knowledge-base (dbbact.org)'),
    citation_text=_citation
)


plugin.pipelines.register_function(
    function=enrich_pipeline,
    inputs={'repseqs': FeatureData[Sequence],
            'table': FeatureTable[Frequency]},
    # outputs=[('all_wordcloud', Visualization),
    #          ('diff_ASVs_table', FeatureData[Differential]),
    #          ('diff_ASVs_heatmap', Visualization),
    #          ('enriched_terms_table', FeatureData[Differential]),
    #          ('enriched_terms_barplot', Visualization),
    #          ('enriched_terms_heatmap', Visualization),
    #          ('enriched_term_venn', Visualization)],
    outputs=[
        ('trimmed_table', FeatureTable[Frequency]),
        ('all_wordcloud', Visualization),
        ('diff_asv_table', FeatureData[Differential]),
        # ('diff_asv_heatmap', Visualization),
        ('enriched_terms_table', FeatureData[Differential]),
        ('enriched_annotations_table', FeatureData[Differential]),
        ('enriched_terms_barplot', Visualization),
        ('enriched_annotations_barplot', Visualization),
        # ('enriched_terms_heatmap', Visualization),
        # ('enriched_term_venn', Visualization),
    ],
    parameters={
        'metadata': Metadata,
        'method': Str % Choices('groups', 'correlation'),
        'field': Str,
        'val1': List[Str],
        'val2': List[Str],
        'pair_field': Str,
        'random_seed': Int,
        'sig_threshold': Float,
        'maxid': Int,
        'statistical_test': Str % Choices(_statistical_tests),
        'transform_function': Str % Choices(_transform_functions),
        'permutations': Int,
        'fdr_method': Str % Choices(_fdr_method),
        'attack': Bool,
    },
    input_descriptions={
        'table': 'The feature table to perform differential abundance on (needs to be normalized or rarified)',
        'repseqs': 'The corresponding representative sequences (if not embedded in the table using the --p-no-hashed-feature-ids parameter when denoising)'
    },
    parameter_descriptions={
        'metadata': 'Metadata (mapping) file for the table',
        'field': 'The metadata field on which to perform the differential abundance test',
        'pair_field': 'If supplied, perform paired differential abundance testing pairing samples in the pair_field metadata field',
        'val1': 'Metadata values (in field) for samples in group1. Can supply multiple values',
        'val2': 'Metadata values (in field) for samples in group2. Can supply multiple values. If not provided, take all samples not in group1',
        'statistical_test': 'The statistic used for the effect size calculation. "meandiff" is the difference in the means between the two groups, '
                            '"stdmeandiff" is the difference in the means normalized by the standard deviation.',
        'transform_function': 'The transformation to apply to the frequency data (for each feature) before calculating the effect size. options: '
                              '"rankdata": rank each feature across samples, '
                              '"log2data": log2 transform each frequency, '
                              '"normdata": normalize to constant sum for each feature, '
                              '"binarydata": convert to present/absent, '
                              '"none": do not apply any transformation',
        'permutations': 'Number of permutations to calculate (higher number is more exact but slower)',
        'random_seed': 'The random seed to use (for fully reproducible results, since test is permutation based)',
        'fdr_method': 'The multiple hypothesis correction method to use. options: '
                      '"dsfdr": The discrete FDR correction (stronger than BHfdr if there are many low prevalence features) (Jiang et al 2017, https://doi.org/10.1128/mSystems.00092-17), '
                      '"bhfdr": Benhaminy-Hochberg FDR correction, '
                      '"byfdr": Benhaminy-Yekutieli FDR correction, '
                      '"filterBH": Low frequency filtering based correction (see Jiang et al 2017, https://doi.org/10.1128/mSystems.00092-17).',
        'method': '"groups" to compare term enrichemnent between significantly enriched sequences in both directions, "correlation" to detect dbbact terms significanly correlated/anti-correlated with the effect size.',
        'maxid': 'The maximal dbBact annotation id to use (to enable replication of results after new annotations are added to dbBact',
        'attack': 'Attack mode',
    },
    output_descriptions={'all_wordcloud': 'Wordcloud of all the features in the input table',
                         'diff_asv_table': 'Table of the differentially abundant ASVs between to the groups acoording to metadata field',
                         'trimmed_table': 'The feature table after trimming of the primers (for alignment to dbBact sequences',
                         'enriched_terms_table': 'dbBact terms enriched in either of the two ASV groups (identified in trimmed_table)',
                         'enriched_terms_barplot': 'barplot of the dbBact terms listed in enriched_terms_table',
                         'enriched_annotations_table': 'dbBact annotations enriched in either of the two ASV groups (identified in trimmed_table)',
                         'enriched_annotations_barplot': 'barplot of the dbBact annotations listed in enriched_anno_table',
                         },
    name='dbBact term enrichment for differential abundance results',
    description=("Identify dbBact terms enriched in results of differential abundance (terms significantly more represented in either of the differential abundance groups or correlated with effect size)")
)


plugin.methods.register_function(
    function=enrichment,
    inputs={'diff': FeatureData[Differential],
            'repseqs': FeatureData[Sequence],
            },
    outputs=[('enriched', FeatureData[Differential])],
    parameters={
        'source': Str % Choices(['dsfdr', 'aldex2', 'dacomp', 'songbird', 'ancom', 'tsv']),
        'method': Str % Choices('groups', 'correlation'),
        'term_type': Str % Choices(['term', 'annotation']),
        'random_seed': Int,
        'sig_threshold': Float,
        'diff_tsv': Str,
        'ancom_stat': Str,
        'maxid': Int,
        'attack': Bool,
    },
    input_descriptions={
        'diff': 'Result of qiime2 differential abundance plugin (using dsfdr/aldex2/dacomp/songbird). Should be .qza. If using ancom/generic tsv, use --p-diff_tsv parameter instead.',
    },
    parameter_descriptions={
        'source': 'Origin of the differentail abundance',
        'method': '"groups" to compare term enrichemnent between significantly enriched sequences in both directions, "correlation" to detect dbbact terms significanly correlated/anti-correlated with the effect size.',
        'term_type': '"term" to get enriched dbBact terms, "annotation" to get list of enriched dbBact annotations in one group compared to the other',
        'random_seed': 'If provided, use as the random seed for the enrichment permutation test (to ensure complete replication)',
        'diff_tsv': ('A tsv table input file (e.g. from ancom when using --p-source ancom, or general tsv when using --p-source tsv).'
                     ' Use instead of --i-diff. When using tsv, file should contain the columns:"id" (sequence), "effect", "pval", "reject".'),
        'ancom_stat': 'the ancom statitical results output file for significant sequence identification (optional - overrides sig-threshold)',
        'maxid': 'The maximal dbBact annotation id to use (to enable replication of results after new annotations are added to dbBact',
        'attack': 'Attack mode',
    },
    output_descriptions={'enriched': 'the enriched dbBact terms'},
    name='dbBact term enrichment for differential abundance results',
    description=("Identify dbBact terms enriched in results of differential abundance (terms significantly more represented in either of the differential abundance groups or correlated with effect size)")
)


plugin.methods.register_function(
    function=embed_seqs,
    inputs={'repseqs': FeatureData[Sequence],
            'table': FeatureTable[Frequency]},
    outputs=[('merged', FeatureTable[Frequency])],
    parameters={},
    input_descriptions={
        'table': 'The biom table with hashed ASV IDs',
        'repseqs': 'The corresponding representative sequences'
    },
    output_descriptions={'merged': 'The biom table with embedded representative sequences'},
    name='embed sequences into hashed biom table',
    description=('embed the representative sequences into the biom table')
)


plugin.methods.register_function(
    function=trim_primers,
    inputs={'repseqs': FeatureData[Sequence],
            'table': FeatureTable[Frequency]},
    outputs=[('trimmed', FeatureTable[Frequency])],
    parameters={},
    input_descriptions={
        'table': 'The biom table with ASVs to be trimmed',
        'repseqs': 'The corresponding representative sequences (if not embedded in the table using the --p-no-hashed-feature-ids parameter when denoising)'
    },
    output_descriptions={'trimmed': 'The biom table with primer-trimmed sequences'},
    name='trim primers from table sequences',
    description=('Make sequences in the biom table compatible with dbBact. The method will automatically identify the dbBact supported primers in the sequences and trim the sequences so '
                 'all sequences in the biom table start at the end of the supported forward primer. Currently supported dbBact primers are V1 (AGAGTTTGATC[AC]TGG[CT]TCAG), '
                 'V3 (CCTACGGG[ACGT][CGT]GC[AT][CG]CAG) and V4 (GTGCCAGC[AC]GCCGCGGTAA)')
)


plugin.methods.register_function(
    function=single_enrichment,
    inputs={'repseqs': FeatureData[Sequence],
            'bg_seqs': FeatureData[Sequence],
            'bg_table': FeatureTable[Frequency],
            'test_seqs': FeatureData[Sequence],
            'test_table': FeatureTable[Frequency]},
    outputs=[('enriched', FeatureData[Differential])],
    parameters={},
    input_descriptions={
        'bg_table': 'The table for background sequences (use this or bg_seqs)',
        'bg_seqs': 'The background sequences (use this or bg_table)',
        'test_table': 'The table with sequences to test for term enrichment (use this or test_seqs)',
        'test_seqs': 'The sequences to test for term enrichment (use this or test_table)',
        'repseqs': 'The corresponding representative sequences'
    },
    output_descriptions={'enriched': 'Table containing the enriched dbBact terms in the test compared to the background table'},
    name='dbBact term enrichment compared to experiment sequence background',
    description=('Test dbBact term enrichment in a test set of sequences compared to an experiment of background sequences')
)


plugin.methods.register_function(
    function=bg_term_enrichment,
    inputs={'repseqs': FeatureData[Sequence],
            'test_seqs': FeatureData[Sequence],
            'test_table': FeatureTable[Frequency]},
    outputs=[('enriched', FeatureData[Differential])],
    parameters={'bg_terms': List[Str],
                'alpha': Float,
                'min_appearance': Int},
    input_descriptions={
        'test_table': 'The table with sequences to test for term enrichment (use this or test_seqs)',
        'test_seqs': 'The sequences to test for term enrichment (use this or test_table)',
        'repseqs': 'The corresponding representative sequences'
    },
    parameter_descriptions={
        'bg_terms': 'List of dbBact terms. Input sequences are compared to sequences from all dbBact experiments containing all the terms. '
                    'For example to compare input sequences to human fecal bacteria from china, use --bg-terms feces china "homo spaiens"',
        'alpha': 'The dsFDR threshold for significantly enriched terms',
        'min_appearance': 'Minimal number of experiments each dbBact background term should appear in (as COMMON)'
    },
    output_descriptions={'enriched': 'Table containing the enriched dbBact terms in the test compared to the background table'},
    name='dbBact term enrichment compared to dbBact sequences associated with a list of terms',
    description=('Test dbBact term enrichment in a test set of sequences compared to all dbBact sequences associated (COMMON) with a set of terms')
)


plugin.methods.register_function(
    function=diff_abundance,
    inputs={'repseqs': FeatureData[Sequence],
            'table': FeatureTable[Frequency]},
    outputs=[('diff', FeatureData[Differential])],
    parameters={
        'metadata': Metadata,
        'field': Str,
        'pair_field': Str,
        'statistical_test': Str % Choices(_statistical_tests),
        'transform_function': Str % Choices(_transform_functions),
        'permutations': Int,
        'alpha': Float,
        'random_seed': Int,
        'fdr_method': Str % Choices(_fdr_method),
        'val1': List[Str],
        'val2': List[Str]
    },
    input_descriptions={
        'table': 'The feature table to perform differential abundance on (needs to be normalized or rarified)',
        'repseqs': 'The corresponding representative sequences (if not embedded in the table using the --p-no-hashed-feature-ids parameter when denoising)'
    },
    parameter_descriptions={
        'alpha': 'The dsFDR threshold for significantly enriched terms',
        'metadata': 'Metadata (mapping) file for the table',
        'field': 'The metadata field on which to perform the differential abundance test',
        'pair_field': 'If supplied, perform paired differential abundance testing pairing samples in the pair_field metadata field',
        'val1': 'Metadata values (in field) for samples in group1. Can supply multiple values',
        'val2': 'Metadata values (in field) for samples in group2. Can supply multiple values. If not provided, take all samples not in group1',
        'statistical_test': 'The statistic used for the effect size calculation. "meandiff" is the difference in the means between the two groups, '
                            '"stdmeandiff" is the difference in the means normalized by the standard deviation.',
        'transform_function': 'The transformation to apply to the frequency data (for each feature) before calculating the effect size. options: '
                              '"rankdata": rank each feature across samples, '
                              '"log2data": log2 transform each frequency, '
                              '"normdata": normalize to constant sum for each feature, '
                              '"binarydata": convert to present/absent, '
                              '"none": do not apply any transformation',
        'permutations': 'Number of permutations to calculate (higher number is more exact but slower)',
        'alpha': 'the FDR threshold for rejecting the null hypothesis',
        'random_seed': 'The random seed to use (for fully reproducible results, since test is permutation based)',
        'fdr_method': 'The multiple hypothesis correction method to use. options: '
                      '"dsfdr": The discrete FDR correction (stronger than BHfdr if there are many low prevalence features) (Jiang et al 2017, https://doi.org/10.1128/mSystems.00092-17), '
                      '"bhfdr": Benhaminy-Hochberg FDR correction, '
                      '"byfdr": Benhaminy-Yekutieli FDR correction, '
                      '"filterBH": Low frequency filtering based correction (see Jiang et al 2017, https://doi.org/10.1128/mSystems.00092-17).'
    },
    output_descriptions={'diff': 'Table of differentially abundant bacteria'},
    name='Calour based differential abundance',
    description=('Identify differentially abundance bacteria between two groups using the calour diff_abundance method')
)


plugin.visualizers.register_function(
    function=draw_wordcloud_vis,
    inputs={'table': FeatureTable[Frequency],
            'repseqs': FeatureData[Sequence],
            },
    parameters={
        'prev_thresh': Float,
        'focus_terms': List[Str],
    },
    input_descriptions={
        'table': 'The biom table to draw the wordcloud for.'
    },
    parameter_descriptions={
        'prev_thresh': 'Mininal prevalence (fraction of samples sequence is present) in order to include sequence in wordcloud stats.',
        'focus_terms': 'show only terms from annotations containing all these terms (comma separated).'
    },
    name='wordcloud',
    description=('draw the dbBact terms wordcloud for ASVs in a given biom table')
)


plugin.visualizers.register_function(
    function=plot_enrichment,
    inputs={'enriched': FeatureData[Differential]},
    parameters={
        'max_show': Int,
        'max_len': Int,
        'labels': List[Str],
        'colors': List[Str],
        'enriched_exp_color': Str
    },
    input_descriptions={
        'enriched': 'The enriched dbbact terms (result of "qiime dbbact enrichment")'
    },
    parameter_descriptions={
        'max_show': 'The maximal number of enriched dbBact terms to show.',
        'max_len': 'Maximal per-term string length (i.e. trim terms after XXX letters for better visibility)',
        'labels': 'Labels of the two groups',
        'colors': 'Color names for group1, group2 bars (valid matplotlib color names)',
        'enriched_exp_color': 'Color name for the enriched experiments number within each bar (if available)'
    },
    name='Visualize dbbact enrichment results (following "enrichment" command)',
    description=('Visualization for qiime dbbact enriched output')
)


plugin.visualizers.register_function(
    function=heatmap,
    inputs={'table': FeatureTable[Frequency],
            'repseqs': FeatureData[Sequence],
            'taxonomy': FeatureData[Taxonomy],
            },
    parameters={'metadata': Metadata,
                'sort_field': Str,
                'cluster': Bool,
                'normalize': Bool,
                'min_abundance': Float,
                },
    input_descriptions={
        'table': 'The table to plot',
        'repseqs': 'If input table contains hashes instead of sequences, supply the representative sequences file used in the table creation.',
        'taxonomy': 'Optional taxonomy file for showing the sequence taxonomy in the heatmap'
    },
    parameter_descriptions={
        'metadata': 'The metadata table (tsv)',
        'sort_field': 'field to sort the samples by before plotting',
        'cluster': 'if true, cluster (reorder) the ASVs before plotting (based on similar behavior across samples',
        'normalize': 'it true, normalize (TSS) the number of reads per sample',
        'min_abundance': 'filter away seqiuences with < min_abundance total reads (over all samples)',
    },
    name='Plot interactive heatmap with dbbact annotations per sequence',
    description=('Plot interactive html heatmap. Samples are columns and sequences are rows.')
)


plugin.visualizers.register_function(
    function=venn,
    inputs={'diff': FeatureData[Differential],
            'repseqs': FeatureData[Sequence],
            },
    parameters={
        'source': Str % Choices(['dsfdr', 'aldex2', 'dacomp', 'songbird', 'ancom', 'tsv']),
        'sig_threshold': Float,
        'diff_tsv': Str,
        'ancom_stat': Str,
        'terms': List[Str],
        'label1': Str,
        'label2': Str,
        'set_colors': List[Str],
        'max_size': Int
    },
    input_descriptions={
        'diff': 'Result of qiime2 differential abundance plugin (using dsfdr/aldex2/dacomp/songbird). Should be .qza. If using ancom/generic tsv, use --p-diff_tsv parameter instead.',
    },
    parameter_descriptions={
        'source': 'Origin of the differentail abundance',
        'diff_tsv': ('A tsv table input file (e.g. from ancom when using --p-source ancom, or general tsv when using --p-source tsv).'
                     ' Use instead of --i-diff. When using tsv, file should contain the columns:"id" (sequence), "effect", "pval", "reject".'),
        'ancom_stat': 'the ancom statitical results output file for significant sequence identification (optional - overrides sig-threshold)',
        'terms': 'List of terms to plot the venn diagram for (blue circle includes sequences appearing in dbBact annotations that include all terms)',
    },
    name='Plot dbBact Venn diagram',
    description='Plot Venn diagram for intersection between two seqs of sequences (from diff. abundance) and dbBact sequences associated with a set of terms',
)



from q2_types.feature_data import DNAFASTAFormat
import biom
import pandas as pd
import numpy as np
import calour as ca
import dbbact_calour
import dbbact_calour.dbbact
from typing import List

from .utils import _iter_fasta, _seqs_from_repseqs, _load_diff_abundance, _test_exact_region, _test_embedded_primers, _trim_known_primer


def embed_seqs(table: biom.Table, repseqs: DNAFASTAFormat) -> biom.Table:
    ids = set(table.ids(axis='observation'))
    id_map = {}
    for chead, cseq in _iter_fasta(str(repseqs)):
        if chead in ids:
            id_map[chead] = cseq
    table.update_ids(id_map, axis='observation', strict=False, inplace=True)
    return table


def enrichment(diff: pd.DataFrame = None, repseqs: DNAFASTAFormat = None, diff_tsv: str = None, source: str = 'dsfdr', method: str = 'groups', term_type: str = 'term', sig_threshold: float = 0.1, ancom_stat: str = None, attack: bool = False, maxid: int = None, random_seed: int = None) -> pd.DataFrame:
    ca.set_log_level('INFO')
    db = dbbact_calour.dbbact.DBBact()
    db.set_log_level('INFO')

    # load the differetial abundance table into a calour experiment
    exp = _load_diff_abundance(diff=diff, diff_tsv=diff_tsv, source=source, ancom_stat=ancom_stat, repseqs=repseqs, sig_threshold=sig_threshold)
    ndata = exp.feature_metadata

    # check if the experiment contains trimmed sequences. otherwise let the user know
    seqs = ndata.sample(n=np.min([500, len(ndata)])).index.values
    region = _test_exact_region(seqs)
    if region is None:
        raise ValueError('Table seems to contain untrimmed sequences. Please run qiime dbbact trim-primers on the table prior to running enrichment.')
    else:
        print('Identified region %s' % region)

    # do the dbbact term enrichment test (using 2 feature groups or feature effect size rank correlation)
    if method == 'groups':
        print('%d ASVs' % len(exp.feature_metadata))
        exp = exp.filter_by_metadata('reject', ['1'], axis='f')
        print('%d significant ASVs' % len(exp.feature_metadata))
        pos_features = ndata[ndata['effect'] > 0].index.values
        res = db.enrichment(exp, features=pos_features, term_type=term_type, max_id=maxid, random_seed=random_seed)
        df = res[0]
    elif method == 'correlation':
        print('searching for dbbact term enrichment for %d ASVs' % len(exp.feature_metadata))
        dd = db.rank_enrichment(exp, 'effect', max_id=maxid, random_seed=random_seed)
        print('found %d enriched terms' % len(dd.feature_metadata))
        df = dd.feature_metadata
        df.rename({'_calour_stat': 'odif'}, axis='columns', inplace=True)
        df.drop('_calour_direction', axis='columns', inplace=True)
    else:
        raise ValueError('Unsupported method %s' % source)

    # qiime2 plugin stuff:
    # need to rename the index to "id"
    # and cannot contain non-numeric columns
    df.drop('term', axis='columns', inplace=True)
    df.index.rename('id', inplace=True)
    return df


def single_enrichment(bg_table: biom.Table = None, test_table: biom.Table = None, bg_seqs: DNAFASTAFormat = None, test_seqs: DNAFASTAFormat = None,
                      repseqs: DNAFASTAFormat = None) -> pd.DataFrame:
    # test we have the correct inputs
    if bg_table is None and bg_seqs is None:
        raise ValueError('Must supply either bg_table or bg_seqs')
    if test_table is None and test_seqs is None:
        raise ValueError('Must supply either test_table or test_seqs')
    if bg_table is not None and bg_seqs is not None:
        raise ValueError('Cannot use both bg_table and bg_seqs')
    if test_table is not None and test_seqs is not None:
        raise ValueError('Cannot use both test_table and test_seqs')

    # load background sequences to bg set
    if bg_table is not None:
        print('using background sequences from biom table')
        bg = set(bg_table.ids(axis='observation'))
    else:
        print('using backgound sequences from fasta file')
        bg = set()
        for chead, cseq in _iter_fasta(str(bg_seqs)):
            bg.add(cseq)
    print('loaded %d background sequences' % len(bg))

    # load test sequences to seqs set
    if test_table is not None:
        print('using test sequences from biom table')
        seqs = set(test_table.ids(axis='observation'))
    else:
        print('using test sequences from fasta file')
        seqs = set()
        for chead, cseq in _iter_fasta(str(test_seqs)):
            seqs.add(cseq)
    print('loaded %d test sequences' % len(seqs))

    db = dbbact_calour.dbbact.DBBact()
    db.set_log_level('INFO')

    # create a mock experiment for carrying the features
    all_features = bg.union(seqs)
    exp = ca.AmpliconExperiment(data=np.ones([1, len(all_features)]),
                                sample_metadata=pd.DataFrame({'_sample_id': ['s1']}),
                                feature_metadata=pd.DataFrame({'_feature_id': all_features}, index=all_features),
                                sparse=False)

    # check if the experiment contains trimmed sequences. otherwise let the user know
    seqs = exp.feature_metadata.sample(n=np.min([500, len(exp.feature_metadata)])).index.values
    region = _test_exact_region(seqs)
    if region is None:
        raise ValueError('Table seems to contain untrimmed sequences. Please run qiime dbbact trim-primers on the table prior to running enrichment.')
    else:
        print('Identified region %s' % region)

    res = db.enrichment(exp=exp, features=seqs, method='card_mean')
    term_table = res[0]
    if len(term_table) == 0:
        print('No significantly enriched terms found')
    else:
        print('%d Significantly enriched terms found' % len(term_table))
    return term_table


def bg_term_enrichment(bg_terms: List[str], test_table: biom.Table = None, test_seqs: DNAFASTAFormat = None,
                       repseqs: DNAFASTAFormat = None, alpha: float = 0.1, min_appearance: int = 2) -> pd.DataFrame:
    print('term_bg_enrichment for %d terms: %s' % (len(bg_terms), bg_terms))

    # test we have the correct inputs
    if test_table is None and test_seqs is None:
        raise ValueError('Must supply either test_table or test_seqs')
    if test_table is not None and test_seqs is not None:
        raise ValueError('Cannot use both test_table and test_seqs')

    # load test sequences to seqs set
    if test_table is not None:
        print('using test sequences from biom table')
        seqs = set(test_table.ids(axis='observation'))
    else:
        print('using test sequences from fasta file')
        seqs = set()
        for chead, cseq in _iter_fasta(str(test_seqs)):
            seqs.add(cseq)
    print('loaded %d test sequences' % len(seqs))

    db = dbbact_calour.dbbact.DBBact()
    db.set_log_level('INFO')

    # prepare the mock AmpliconExperiment for the background enrichment.
    # We only care about the features (sequencesw)
    exp = ca.AmpliconExperiment(data=np.ones([1, len(seqs)]),
                                sample_metadata=pd.DataFrame({'_sample_id': ['s1']}),
                                feature_metadata=pd.DataFrame({'_feature_id': seqs}, index=seqs),
                                sparse=False)

    # check if the experiment contains trimmed sequences. otherwise let the user know
    seqs = exp.feature_metadata.sample(n=np.min([500, len(exp.feature_metadata)])).index.values
    region = _test_exact_region(seqs)
    if region is None:
        raise ValueError('Table seems to contain untrimmed sequences. Please run qiime dbbact trim-primers on the table prior to running enrichment.')
    else:
        print('Identified region %s' % region)

    exp = exp.normalize()

    # do the background enrichment analysis
    res = db.background_enrich(bg_terms, exp, ignore_exp=None, min_appearance=min_appearance, include_shared=True, alpha=alpha)
    df = res[0]
    df.drop('term', axis='columns', inplace=True)
    df.index.rename('id', inplace=True)

    return res[0]


def diff_abundance(table: biom.Table,
                   metadata: pd.DataFrame,
                   field: str,
                   pair_field: str = None,
                   # metadata2: MetadataColumn = None,
                   repseqs: DNAFASTAFormat = None,
                   statistical_test: str = 'meandiff',
                   transform_function: str = 'rankdata',
                   alpha: float = 0.1,
                   permutations: int = 1000,
                   random_seed: int = 0,
                   fdr_method: str = 'dsfdr',
                   val1: List[str] = None,
                   val2: List[str] = None) -> pd.DataFrame:

    # allow debug info. q2 takes care of what to show using the --verbose flag
    ca.set_log_level('INFO')

    # set random seed to None (i.e. random every time) if 0 is provided
    if random_seed == 0:
        random_seed = None

    metadata = metadata.to_dataframe()

    # take care of the 'none' option for the transfrom
    if transform_function == 'none':
        transform_function = None

    # check and prepare the metadata columns for the testing
    md_fields = metadata.columns
    if field not in md_fields:
        raise ValueError('Field %s not found in metadata file' % field)
    if pair_field is not None:
        if pair_field not in md_fields:
            raise ValueError('Pair field %s not found in metadata file' % field)
    labels = metadata[field].unique()
    n = len(labels)
    if n == 1:
        raise ValueError('Only one category in metadata column. Aborting')
    if val1 is None and val2 is None:
        if n == 2:
            val1 = [labels[0]]
            val2 = [labels[1]]
            print('No values supplied for field, assigning from the two values present (%s, %s)' % (val1, val2))
        else:
            raise ValueError('Cannot perform %s test on more than two categories in metadata and val1 and val2 not supplied. Aborting' % statistical_test)

    if val1 is not None:
        if len([x for x in metadata[field] if x in val1]) == 0:
            raise ValueError('No values matching val1 (%s) found in metadata field %s' % (val1, field))
    if val2 is not None:
        if len([x for x in metadata[field] if x in val2]) == 0:
            raise ValueError('No values matching val2 (%s) found in metadata field %s' % (val2, field))

    # create the calour.AmpliconExperiment from the table and metadata
    samples = table.ids(axis='sample')
    features = table.ids(axis='observation')
    feature_metadata = pd.DataFrame(index=features, columns={'_feature_id': features})
    # use repseqs if supplied
    if repseqs is not None:
        print('converting hashes to sequences using repseqs file')
        feature_metadata = _seqs_from_repseqs(feature_metadata, repseqs)
    metadata = metadata.filter(items=samples, axis='index')
    exp = ca.AmpliconExperiment(data=table.transpose().matrix_data, sample_metadata=metadata, feature_metadata=feature_metadata, sparse=False)
    print('created experiment %r' % exp)

    if pair_field is None:
        print('performing differential abundance testing')
        dd = exp.diff_abundance(field, val1, val2, alpha=alpha, random_seed=random_seed, numperm=permutations, transform=transform_function, method=statistical_test, fdr_method=fdr_method)
    else:
        print('performing paired differential abundance testing')
        dd = exp.diff_abundance_paired(pair_field, field, val1, val2, alpha=alpha, random_seed=random_seed, numperm=permutations, transform=transform_function, method=statistical_test, fdr_method=fdr_method)

    # if no significant results found, need to raise value error since qiime2 does not support empty diff abundance results
    if len(dd.feature_metadata) == 0:
        raise ValueError('No significant results found')

    res = pd.DataFrame({'Reject': 1,
                        'Statistic': dd.feature_metadata['_calour_stat'],
                        'raw pvalue': dd.feature_metadata['_calour_pval'],
                        'dir': dd.feature_metadata['_calour_stat'] > 0,
                        'qvalue': dd.feature_metadata['_calour_qval']},
                       index=dd.feature_metadata.index)
    res['dir'] = res['dir'].astype(int)
    res.index.rename('featureid', inplace=True)
    return res


def enrich_pipeline(ctx,
                    table,
                    # output_dir: str,
                    metadata,
                    field,
                    pair_field=None,
                    repseqs=None,
                    statistical_test='meandiff',
                    transform_function='rankdata',
                    permutations=1000,
                    fdr_method='dsfdr',
                    val1=None,
                    val2=None,
                    method='groups',
                    sig_threshold=0.1,
                    attack=False,
                    maxid=None,
                    random_seed=0):
    res = []

    print('trimming primers if needed')
    trim_func = ctx.get_action('dbbact', 'trim_primers')
    table, = trim_func(table=table, repseqs=repseqs)
    res.append(table)

    print('generating initial wordcloud')
    wordcloud_func = ctx.get_action('dbbact', 'draw_wordcloud_vis')
    wordcloud, = wordcloud_func(table=table, repseqs=repseqs)
    res.append(wordcloud)

    print('detecting differetially abundant features')
    diff_func = ctx.get_action('dbbact', 'diff_abundance')
    diff_table, = diff_func(table=table, metadata=metadata, field=field, pair_field=pair_field, repseqs=repseqs, statistical_test=statistical_test,
                            transform_function=transform_function, alpha=sig_threshold, permutations=permutations, random_seed=random_seed, fdr_method=fdr_method,
                            val1=val1, val2=val2)
    diff_table_df = diff_table.view(pd.DataFrame)
    if len(diff_table_df) == 0:
        raise ValueError('No significant ASVs found to differentiate between %s and %s in field %s' % (val1, val2, field))
    res.append(diff_table)

    # print('creating diff ASV heatmap')
    # heatmap_func = ctx.get_action('dbbact', 'heatmap')
    # diff_heatmap = heatmap_func(table=diff_table, metadata=metadata, sort_field=field, cluster=False, repseqs=repseqs)
    # res.append(diff_heatmap)

    print('detecting enriched dbBact terms')
    enrich_func = ctx.get_action('dbbact', 'enrichment')
    enriched, = enrich_func(diff=diff_table, source='dsfdr', method=method, sig_threshold=sig_threshold, attack=attack, maxid=maxid, random_seed=random_seed)
    enriched_df = enriched.view(pd.DataFrame)
    print('found %d enriched dbbact terms' % len(enriched_df))
    res.append(enriched)

    print('detecting enriched dbBact annotations')
    enriched_anno, = enrich_func(diff=diff_table, source='dsfdr', term_type='annotation', method=method, sig_threshold=sig_threshold, attack=attack, maxid=maxid, random_seed=random_seed)
    enriched_anno_df = enriched.view(pd.DataFrame)
    print('found %d enriched dbbact annotations' % len(enriched_anno_df))
    res.append(enriched_anno)

    print('creating enriched terms barplot')
    # create the lables for the 2 groups
    metadata_df = metadata.to_dataframe()
    if val1 is not None:
        label1 = ",".join(val1)
    if val2 is not None:
        label2 = ",".join(val2)
    if val1 is None and val2 is None:
        uvals = metadata_df[field].unique()
        label1 = uvals[0]
        label2 = uvals[1]
    elif val2 is None:
        label2 = 'NOT ' + label1

    terms_barplot_func = ctx.get_action('dbbact', 'plot_enrichment')
    enriched_barplot, = terms_barplot_func(enriched=enriched, labels=[label1, label2])
    res.append(enriched_barplot)

    print('creating enriched annotations barplot')
    enriched_annotations_barplot, = terms_barplot_func(enriched=enriched_anno, labels=[label1, label2])
    res.append(enriched_annotations_barplot)

    print('done')
    return tuple(res)


def trim_primers(table: biom.Table, repseqs: DNAFASTAFormat = None) -> biom.Table:
    '''Remove dbBact supported primers from all sequences
    This is done since all sequences in dbBact are indexed based on the sequence immediately following the supported primers (v1/v3/v4)

    The primer is detected in several ways:
    1.

    Parameters
    ----------
    table: biom.Tabls
        Can contain either sequences (i.e. dada2/deblur was run with --p-no-hashed-feature-ids) or hashes (i.e. dada2/deblur was run with --p-hashed-feature-ids, which is the default for qiime2)
        If the table contains hashes, you must supplu the repseqs parameter as well
    repseqs: the hash->sequence conversion file
        The output of denoising stored to the --o-representative-sequences file
    '''
    print('trim-primers')

    # incorporate the repseqs into the biom table if needed
    if repseqs is not None:
        print('converting hashes to sequences using repseqs file')
        table = embed_seqs(table, repseqs)

    seqs = table.ids(axis='observation')

    # check if table contains hashed sequences and repseqs not supplied
    if len(seqs[0]) == 32:
        if repseqs is None:
            raise ValueError('Table seems to contain hashes and not sequences. Please supply the --i-repseqs representative sequences file, or create the table (deblur/dada2) with the --p-no-hashed-feature-ids flag.')

    # take a random subset of ASVs to examine
    num_rand = np.min([100, len(seqs)])
    test_seqs = np.random.choice(seqs, num_rand, replace=False)

    # look if there is an exact match to any dbBact primer region
    region = _test_exact_region(test_seqs, min_fraction=0.5, ltrim=0)
    if region is not None:
        print('Found exact match for region %s. No need for trimming.' % region)
    else:
        # no exact match - so maybe primer is embedded?
        min_primer_len = 10
        # look if we find a supported primer (v1/v3/v4) in the first 25 nucleotides for at least 0.25 of the sequences
        primer, region = _test_embedded_primers(seqs, max_start=25, min_primer_len=min_primer_len, min_fraction=0.25)
        if region is not None:
            print('found embedded primer %s for region %s. removing it.' % (primer, region))
            # let's trim the primers
            trimmed_seqs_map = _trim_known_primer(seqs, primer, rprimer=None, length=0, remove_ambig=True, keep_primers=False)
        else:
            # no primer embedded. So maybe need to trim just first few bases? (i.e. less than min_primer_len of the primer is in the sequences)
            region = None
            for cpos in range(min_primer_len):
                region = _test_exact_region(test_seqs, min_fraction=0.5, ltrim=cpos)
                if region is not None:
                    break
            if region is not None:
                # found a matching region - lets trim it
                print('found exact region %s after %d left trim' % (region, cpos))
                trimmed_seqs_map = {}
                for cseq in seqs:
                    trimmed_seqs_map[cseq] = cseq[cpos:]
            else:
                msg = 'No matching primers found. First 10 sequences used for testing (out of 100) are: \n'
                msg += '%s\n' % test_seqs[:min(10, len(test_seqs))]
                msg += 'dbBact supports only V1 (AGAGTTTGATC[AC]TGG[CT]TCAG), V3 (CCTACGGG[ACGT][CGT]GC[AT][CG]CAG) or V4 (GTGCCAGC[AC]GCCGCGGTAA) primers. A few possible error sources: \n' \
                       '1. Are your reads from these regions?\n' \
                       '2. Are yourreads reverse-complemented?\n' \
                       '3. Did you trim a part of the reads following the primer? Reads should start immediately after the end of the primer\n'
                raise ValueError(msg)
        keep_ids = trimmed_seqs_map.keys()
        print('keeping %d out of %d sequences' % (len(keep_ids), len(seqs)))

        # check if we got identical sequences originating in reads with different primers (should not happen - could be due to prior incomplete primer trimming)
        if len(keep_ids) != len(set(trimmed_seqs_map.values())):
            print('Duplicate sequences identified after primer removal. This could be due to using an external primer trimming software (such as cutadapt). For cutadapt, we recommend using the "--discard-untrimmed" parameter.')
            rev_dict = {}
            for key, value in trimmed_seqs_map.items():
                rev_dict.setdefault(value, set()).add(key)
            result = filter(lambda x: len(x) > 1, rev_dict.values())
            print('The following sequences are identical after primer removal (showing first 10 sequences):')
            result = list(result)
            if len(result) > 10:
                result = result[:10]
            print(result)

            # collapse the reads that have same sequence after primer trimming
            def bin_f(cid, cm):
                return trimmed_seqs_map[cid]
            print('Collapsing on sequences after trim')
            table = table.collapse(bin_f, axis='observation')

        table.filter(keep_ids, axis='observation', inplace=True)
        table.update_ids(trimmed_seqs_map, axis='observation', strict=False, inplace=True)

    return table




from setuptools import setup, find_packages

import re
import ast


# version parsing from __init__ pulled from Flask's setup.py
# https://github.com/mitsuhiko/flask/blob/master/setup.py
_version_re = re.compile(r'__version__\s+=\s+(.*)')

with open('q2_dbbact/__init__.py', 'rb') as f:
    hit = _version_re.search(f.read().decode('utf-8')).group(1)
    version = str(ast.literal_eval(hit))

classifiers = [
    'Development Status :: 4 - Beta',
    'License :: OSI Approved :: BSD License',
    'Environment :: Plugins',
    'Topic :: Software Development :: Libraries :: Application Frameworks',
    'Topic :: Scientific/Engineering',
    'Topic :: Scientific/Engineering :: Bio-Informatics',
    'Programming Language :: Python',
    'Programming Language :: Python :: 3',
    'Programming Language :: Python :: 3.6',
    'Programming Language :: Python :: 3.7',
    'Programming Language :: Python :: 3.8',
    'Operating System :: Unix',
    'Operating System :: POSIX',
    'Operating System :: MacOS :: MacOS X',
    'Operating System :: Microsoft :: Windows']


description = 'A qiime2 (https://qiime2.org/) plugin for dbBact (http://dbbact.org) annotations of microbiome experiments'

with open('README.md') as f:
    long_description = f.read()

keywords = 'microbiome qiime2 dbbact database analysis bioinformatics',

setup(
    name="q2-dbbact",
    version=version,
    packages=find_packages(),
    license='BSD',
    long_description=long_description,
    long_description_content_type="text/markdown",
    keywords=keywords,
    classifiers=classifiers,
    author="dbBact team",
    author_email="info@dbbact.org",
    url="http://dbbact.org",
    description=description,
    python_requires='>=3.6',
    entry_points={
        "qiime2.plugins":
        ["q2-dbbact=q2_dbbact.plugin_setup:plugin"]
    },
    install_requires=[
        'wordcloud',
        'calour',
        'dbbact-calour'
    ]
)



from q2_types.feature_data import DNAFASTAFormat
import biom
import pandas as pd
import numpy as np
import os
import calour as ca
import dbbact_calour
import dbbact_calour.dbbact
from typing import List

from .utils import _seqs_from_repseqs, _load_diff_abundance, _test_exact_region


def draw_wordcloud_vis(output_dir: str, table: biom.Table, repseqs: DNAFASTAFormat = None, prev_thresh: float = 0.3, focus_terms: List[str] = None):
    '''draw the wordcloud for features in the biom table and save outputs'''
    db = dbbact_calour.dbbact.DBBact()
    db.set_log_level('INFO')

    df = table.to_dataframe()

    if repseqs is not None:
        df = _seqs_from_repseqs(data=df, repseqs=repseqs)

    if len(df.iloc[0].name) == 32:
        raise ValueError('input table seems to contain hashes and not sequences. Please supply the rep-seqs.qza file.')

    exp = ca.AmpliconExperiment.from_pandas(df.T)

    # check if the experiment contains trimmed sequences. otherwise let the user know
    seqs = exp.feature_metadata.sample(n=np.min([500, len(exp.feature_metadata)])).index.values
    region = _test_exact_region(seqs)
    if region is None:
        raise ValueError('Table seems to contain untrimmed sequences. Please run qiime dbbact trim-primers on the table prior to running enrichment.')
    else:
        print('Identified region %s' % region)

    print('%d ASVs in table before prevalence filtering' % len(exp.feature_metadata))
    exp = exp.filter_prevalence(prev_thresh)
    print('%d ASVs in table remain after prevalence filtering (%f)' % (len(exp.feature_metadata), prev_thresh))
    if focus_terms is not None:
        print('Using %d focus terms: %s' % (len(focus_terms), focus_terms))
    print('getting term stats')
    fscores, recall, precision, term_count, reduced_f = db.get_wordcloud_stats(exp=exp, focus_terms=focus_terms)
    df = pd.DataFrame(data={'fscore': fscores, 'recall': recall, 'precision': precision, 'term_count': term_count, 'reduced_f': reduced_f})
    df.to_csv(os.path.join(output_dir, 'scores.tsv'), sep='\t')
    print('drawing wordcloud')
    f = db.draw_wordcloud(exp, focus_terms=focus_terms)
    f.savefig(os.path.join(output_dir, 'wordcloud.pdf'))
    f.savefig(os.path.join(output_dir, 'wordcloud.svg'))
    with open(os.path.join(output_dir, 'index.html'), 'w') as fl:
        fl.write('<html><body>\n')
        fl.write('<h1>dbBact wordcloud</h1>\n')
        fl.write('<h2>prevalence threshold=%f</h2>\n' % prev_thresh)
        if focus_terms is not None:
            fl.write('<h2>focus-terms: %s</h2>' % focus_terms)
        fl.write('<br><img src="wordcloud.svg" alt="wordcloud">\n')
        fl.write('<a href="wordcloud.pdf">Download as PDF</a><br>\n')
        fl.write('<a href="scores.tsv">Download scores as tsv</a><br>\n')
        fl.write('</body></html>')


def plot_enrichment(output_dir: str, enriched: pd.DataFrame, max_show: int = 10, max_len: int = 40, colors: List[str] = ['green', 'red'],
                    labels: List[str] = ['group1', 'group2'], enriched_exp_color: str = 'white'):
    '''Plot the enriched terms bar plot
    '''
    enriched['term'] = enriched.index.values
    ax = ca.plotting.plot_enrichment(None, enriched, max_show=max_show, max_len=max_len, colors=colors, labels=labels, enriched_exp_color=enriched_exp_color)
    ax.figure.tight_layout()
    ax.figure.savefig(os.path.join(output_dir, 'enriched_terms.svg'), bbox_inches='tight')
    ax.figure.savefig(os.path.join(output_dir, 'enriched_terms.pdf'), bbox_inches='tight')
    enriched.to_csv(os.path.join(output_dir, 'enriched_terms.tsv'), sep='\t')
    with open(os.path.join(output_dir, 'index.html'), 'w') as fl:
        fl.write('<html><body>\n')
        fl.write('<h1>dbBact enriched terms</h1>\n')
        fl.write('<img src="enriched_terms.svg" alt="enriched terms barplot"><br><br>\n')
        fl.write('<a href="enriched_terms.pdf">Download as PDF</a><br>\n')
        fl.write('<a href="enriched_terms.tsv">Download enriched terms as tsv</a><br>\n')
        fl.write('</body></html>')


def heatmap(output_dir: str, table: biom.Table, metadata: pd.DataFrame, sort_field: str = None, cluster: bool = True, min_abundance: float = 10, normalize: bool = True,
            taxonomy: pd.DataFrame = None, repseqs: DNAFASTAFormat = None):
    ca.set_log_level('INFO')
    print('creating experiment')
    metadata = metadata.to_dataframe()
    samples = table.ids(axis='sample')
    features = table.ids(axis='observation')
    # combine the taxonomy
    feature_metadata = pd.DataFrame(index=features, columns={'_feature_id': features})
    if taxonomy is not None:
        print('adding taxonomy')
        feature_metadata['taxonomy'] = taxonomy['Taxon']
    else:
        feature_metadata['taxonomy'] = 'NA'
    # use repseqs if supplied
    if repseqs is not None:
        print('converting hashes to sequences using repseqs file')
        feature_metadata = _seqs_from_repseqs(feature_metadata, repseqs)

    metadata = metadata.filter(items=samples, axis='index')
    exp = ca.AmpliconExperiment(data=table.transpose().matrix_data, sample_metadata=metadata, feature_metadata=feature_metadata, sparse=False)
    print('created experiment %r' % exp)
    if normalize:
        print('normalizing to 10000 reads/sample')
        exp = exp.normalize(10000)
    if min_abundance > 0:
        print('removing sequences with sum abundance < %f' % min_abundance)
        exp = exp.filter_sum_abundance(min_abundance)
        print('remaining experiment %r' % exp)
    if sort_field is not None:
        print('sorting by field %s' % sort_field)
        exp = exp.sort_samples(sort_field)
    if cluster:
        print('clustering sequences')
        exp = exp.cluster_features(10)
    print('generating heatmap to output file %s' % output_dir)
    exp.export_html(output_file=os.path.join(output_dir, 'index.html'), sample_field=sort_field, feature_field='taxonomy')


def venn(output_dir: str, terms: List[str], diff: pd.DataFrame = None, repseqs: DNAFASTAFormat = None, diff_tsv: str = None, source: str = 'dsfdr', sig_threshold: float = 0.1, ancom_stat: str = None,
         label1: str = 'Negative', label2: str = 'Positive', set_colors: List[str] = ['red', 'green', 'mediumblue'], max_size: int = 1000):
    ca.set_log_level('INFO')
    db = dbbact_calour.dbbact.DBBact()
    db.set_log_level('INFO')

    # load the differetial abundance table into a calour experiment
    exp = _load_diff_abundance(diff=diff, diff_tsv=diff_tsv, source=source, ancom_stat=ancom_stat, repseqs=repseqs, sig_threshold=sig_threshold)
    # exp.feature_metadata['_calour_direction'] = exp.feature_metadata['dir']
    exp.feature_metadata['_calour_direction'] = exp.feature_metadata['dir'].replace({0: label1, 1: label2}, inplace=False)

    # check if the experiment contains trimmed sequences. otherwise let the user know
    seqs = exp.feature_metadata.sample(n=np.min([500, len(exp.feature_metadata)])).index.values
    region = _test_exact_region(seqs)
    if region is None:
        raise ValueError('Table seems to contain untrimmed sequences. Please run qiime dbbact trim-primers on the table prior to running enrichment.')
    else:
        print('Identified region %s' % region)

    f = db.plot_term_venn_all(terms, exp, max_size=max_size, set_colors=set_colors)
    f.savefig(os.path.join(output_dir, 'venn.svg'))
    f.savefig(os.path.join(output_dir, 'venn.pdf'))
    with open(os.path.join(output_dir, 'index.html'), 'w') as fl:
        fl.write('<html><body>\n')
        fl.write('<h1>dbBact venn for terms %s</h1>\n' % ','.join(terms))
        fl.write('<img src="venn.svg" alt="term venn diagram"><br><br>\n')
        fl.write('<a href="venn.pdf">Download as PDF</a><br>\n')
        fl.write('</body></html>')



from collections import defaultdict
import re

from q2_types.feature_data import DNAFASTAFormat
import pandas as pd
import numpy as np

import calour as ca


def _iter_fasta(fp):
    '''Read fasta file into iterator.

    Fasta file must contain header line (starting with ">") and one or more sequence lines.

    Parameters
    ----------
    fp: str
        name of the fasta file

    Yields
    ------
    header: str
        the header line (without ">")
    sequence: str
        the sequence ('ACGT'). Both header and sequence are whitespace stripped.
    '''
    # skip non-header lines at beginning of file
    with open(fp, 'r') as fl:
        for cline in fl:
            if cline[0] == ">":
                title = cline[1:].rstrip()
                break
            print('Fasta file %s has no headers' % fp)
            return

        lines = []
        for cline in fl:
            if cline[0] == ">":
                yield title, ''.join(lines)
                lines = []
                title = cline[1:].strip()
                continue
            lines.append(cline.strip())
        yield title, "".join(lines)


def _seqs_from_repseqs(data: pd.DataFrame, repseqs: DNAFASTAFormat):
    '''add sequences from repseqs to the data

    Parameters
    ----------

    Returns
    -------
    data: pandas.DataFrame
        with the updated index containing sequences instead of hashes
    '''
    # check if feature ids are hashes by looking at first feature
    if len(data.iloc[0].name) == 32:
        print('replacing hashes with sequences')
        rename_ids = {}
        for chead, cseq in _iter_fasta(str(repseqs)):
            if chead in data.index:
                rename_ids[chead] = cseq
        if len(rename_ids) < len(data):
            print('Missing rep. sequences for some hashes. Found %d out of %d' % (len(rename_ids), len(repseqs)))
        data.rename(index=rename_ids, inplace=True)
    else:
        print('table does not seem to contain qiime2 hashes for ASV ids.\nFirst ASV id is %s\n' % data.iloc[0].name)
    return data


def _load_diff_abundance(diff, diff_tsv, source, ancom_stat, repseqs=None, sig_threshold=0.1):
    '''Load a diff_abundance qiime2 artifact anc convert it to a calour experiment

    Supports multiple formats (songbird / aldex2 / ancom / tsv)

    Parameters
    ----------
    diff: biom.Table or None
        the differential abundance biom table (or None)
    diff_tsv: str or None
        A tsv file name containing the diff. abundance table
    source: str
        the source for the diff. abundance output. options are:
            "songbird"
            "aldex2"
            "ancom"
            "tsv": use an external tsv file
    ancom_stat: str, optional
        name of the ancom_stat output file (for source=ancom)
    repseqs: q2_types.feature_data.DNAFASTAFormat or None
        if not None, incorporate representative sequences from this file (instead of the hashes in the table)
    sig_threshold: float
        the significance threshold for including sequences from the diff. abundance test

    Returns
    -------
    ca.AmpliconExperiment
        containing the diff. abundance features and the direction each feature is associated with
    '''
    # handle either diff (qza diffabundance artifact) or diff_tsv(tsv file resulting from diff abundance export i.e. ancom etc.)
    if diff is None and diff_tsv is None:
        raise ValueError('Need to supply either --i-diff or --p-diff-tsv differential results input file (none were supplied)')
    if diff is not None and diff_tsv is not None:
        raise ValueError('Need to supply either --i-diff or --p-diff-tsv differential results input file (both were supplied)')

    # get the differential abundance dataframe. index is sequence
    if diff is not None:
        data = diff
    else:
        try:
            data = pd.read_csv(diff_tsv, sep='\t')
            data.set_index(data.columns[0], drop=False, inplace=True)
        except Exception as e:
            print('failed to read diff_tsv file %s. Is is a valid tsv differential abundance file?' % diff_tsv)
            raise e

    # if needed, replace q2 hashes with sequences from the rep-seqs file
    if repseqs is not None:
        data = _seqs_from_repseqs(data=data, repseqs=repseqs)

    ndata = data.copy()

    # check if feature ids are hashes by looking at first feature
    if len(data.iloc[0].name) == 32:
        print('ASV IDs seem to be hashes. First id=%s' % data.iloc[0].name)
        raise ValueError('Input file contains sequence hashes instead of actual sequences.\nPlease use qiime dbbact enrichment-hash and supply the rep_seqs.qza.')

    # preprocess the dataframe to capture the correct per-feature differential abundance fields
    # these will be translated into "effect" (effect size),"pval" (p-value for differential abundance), "dir" (higher in which group) and "reject" (reject the null hypothesis of similar distribution in both groups)
    ndata['_feature_id'] = ndata.index.values
    if source == 'dsfdr':
        ndata = pd.DataFrame(data={'dir': data['Statistic'] > 0,
                                   'pval': data['raw pvalue'],
                                   'effect': data['Statistic'],
                                   'reject': data['Reject']}, index=ndata.index)
    elif source == 'songbird':
        vals = list(ndata.columns.values)
        if 'Intercept' in vals:
            vals.remove('Intercept')
        else:
            print('"Intercept" not in input data. Colums are: %s. Is it a songbird differentials file?' % vals)
            raise ValueError('File does not seem to be a songbird differentials file (no "Intercept" column)')
        vals.remove('_feature_id')
        if len(vals) > 1:
            print('More than 1 column detected: %s' % vals)
            raise ValueError('More than 1 column in songbird results (%s). Current version works only on 1 column' % vals)

        # we don't have p-values so do not reject any
        ndata = pd.DataFrame(data={'effect': ndata[vals[0]], 'reject': 1}, index=ndata.index)
    elif source == 'aldex2':
        ndata = pd.DataFrame(data={'dir': data['effect'] > 0,
                                   'pval': data['we.eBH'],
                                   'effect': data['effect'],
                                   'reject': data['we.eBH'] < sig_threshold}, index=ndata.index)
        ndata = ndata[ndata['reject'] == 1]
    elif source == 'ancom':
        # if user supplied the ancom stats file, use it to get significant (null hypothesis rejects)
        # otherwise, use the sig_threshold for the W stat
        ndata = pd.DataFrame(data={'dir': data['clr'] > 0,
                                   'effect': data['clr'],
                                   'pval': 0,
                                   'reject': data['W'] > sig_threshold}, index=ndata['id'])
        if ancom_stat is not None:
            try:
                rejects = pd.read_csv(ancom_stat, sep='\t', index_col=0)
                ndata = ndata.join(rejects)
                ndata['reject'] = ndata['Reject null hypothesis'] * 1
            except Exception as e:
                print('failed to read ancom-stat file %s. Is is a valid tsv differential abundance file?' % ancom_stat)
                raise e
    elif source == 'tsv':
        if 'pval' not in data.columns:
            raise ValueError('input tsv file does not contain "pval" column')
        if 'effect' not in data.columns:
            raise ValueError('input tsv file does not contain "effect" column')
        if 'reject' not in data.columns:
            data['reject'] = 1
        ndata = pd.DataFrame(data={'dir': data['effect'] > 0,
                                   'effect': data['effect'],
                                   'pval': data['pval'],
                                   'reject': data['reject']}, index=ndata['id'])
    else:
        raise ValueError('Unsupported source %s' % source)

    # create a new calour experiment using sequences as features and 1 sample (we don't use the sample)
    exp = ca.AmpliconExperiment(data=np.ones([1, len(ndata)]),
                                sample_metadata=pd.DataFrame({'_sample_id': ['s1']}),
                                feature_metadata=ndata,
                                sparse=False)
    return exp


def _test_exact_region(seqs, kmers={'v4': ['TACG'], 'v3': ['TGGG', 'TGAG'], 'v1': ['GACG', 'GATG', 'ATTG']}, min_fraction=0.5, ltrim=0):
    '''Test if a fasta file starts with known region k-mers

    Parameters
    ----------
    seqs: list of str
        the sequences to test
    kmers: dict of {region_name(str): [kmers(str)]}
        the kmers expected to begin each primer region
        get the values using ~/scripts/count_kmer.py -d v1
    num_reads: int
        the maximal number of reads per fileto test
    min_fraction: float
        the minimal expected fraction of reads in the region starting with the kmers (summed over all kmers of the region)
    ltrim: int, optional
        position of first nucleotide to start with. 0 to start from beginning

    Returns
    -------
    region: str or None
        the matching region or None if no region
    '''
    kmer_len = len(kmers['v4'][0])
    print('testing kmer head for region on %d sequences:' % len(seqs))
    kmer_dist = defaultdict(float)
    for cseq in seqs:
        cseq = cseq[ltrim:ltrim + kmer_len]
        for cregion, ckmers in kmers.items():
            if cseq in ckmers:
                kmer_dist[cregion] += 1

    if len(kmer_dist) > 0:
        maxregion = max(kmer_dist, key=kmer_dist.get)
        if kmer_dist[maxregion] / len(seqs) >= min_fraction:
            return maxregion
        else:
            print('No region exact match detected. Maximal region was %s but fraction of matching sequences is %f.' % (maxregion, kmer_dist[maxregion] / len(seqs)))
            return None
    print('No region exact match detected. No sequence matched any of the regions.')
    return None


def _test_embedded_primers(seqs, primers={'AGAGTTTGATC[AC]TGG[CT]TCAG': 'v1', 'CCTACGGG[ACGT][CGT]GC[AT][CG]CAG': 'v3', 'GTGCCAGC[AC]GCCGCGGTAA': 'v4'}, max_start=25, min_primer_len=10, min_fraction=0.25):
    '''Check if the sequences contain one of a given set of primers in the beginning.

    Parameters
    ----------
    seqs: list of str
        the sequences to test
    primers: dict of {primer(str): region_name(str)}, optional
        the primers to test for
    max_start: int, optional
        maximal start position for the primer (i.e. do not return if primer starts after position max_start)
    min_primer_len: int, optional
        trim primers to keep only min_primer_len last chars
    num_reads: int, optional
        the number of reads in the fasta file to process
    min_fraction: float, optional
        need at least min_fraction sequence matches in order to return the primer

    Returns
    -------
    primer: str or None
        the primer identified as maximal (if >min_fraction matches in tested reads)
    primer_name: str or None
        the name of the primer region identified
    '''
    # attach the base_dir if needed
    print('Testing %d sequences for embedded primers (%d)' % (len(seqs), len(primers)))

    # trim the primers if needed
    if min_primer_len is not None:
        print('Trimming primers before test to length %d' % min_primer_len)
        new_primers = {}
        for k, v in primers.items():
            pos = len(k)
            numchars = 0
            newp = ''
            while True:
                if numchars >= min_primer_len:
                    break
                pos = pos - 1
                if pos < 0:
                    break
                if k[pos] != ']':
                    newp = k[pos] + newp
                    numchars += 1
                    continue
                while k[pos] != '[':
                    newp = k[pos] + newp
                    pos = pos - 1
                newp = k[pos] + newp
                numchars += 1
            new_primers[newp] = v
        primers = new_primers
        print('Trimmed primers are: %s' % primers)

    # scan the files
    matches = defaultdict(float)
    for cseq in seqs:
        cseq = cseq.upper()
        for cprimer in primers.keys():
            ccseq = cseq[:max_start + len(cprimer)]
            match = re.search(cprimer, ccseq)
            if match is not None:
                matches[cprimer] += 1

    if len(matches) == 0:
        print('No embedded primers found in sequences')
        return None, None

    max_primer = max(matches, key=matches.get)
    match_fraction = matches[max_primer] / len(seqs)
    print('Found maximal matching primer: %s with enough matches (%f)' % (max_primer, match_fraction))
    if match_fraction < min_fraction:
        print('Not using primer. Maximal matches are not enough.')
        return None, None
    return max_primer, primers[max_primer]


def _trim_known_primer(seqs, fprimer, rprimer=None, length=150, remove_ambig=True, keep_primers=False):
    '''Trim until a known primer in the sequences

    Parameters
    ----------
    seqs: list of Str
        the sequences to trim ('ACGT')
    fprimer: str
        the forward primer to use ('ACGT')
    rprimer: str or None, optional
        the reverse primer to use ('ACGT') or None to not match reverse (trim only according to fprimer and length)
    length: int, optional
        the number of bases after the end of the primer to keep
    remove_ambig: bool, optional
    keep_primers: bool, optional
        If True, keep the primers trim until the beginning of the fprimer). If False, start sequences at the end of fprimer.

    Returns
    -------
    newseqs: dict, where key is the seqs_df index, and value is the trimmed sequence
    NOTE: sequences not containing the primer are not included in the output
    '''
    print('trimming %d sequences with primer %s' % (len(seqs), fprimer))
    newseqs = {}
    for ccseq in seqs:
        cseq = ccseq.upper()
        reverse_primer_seq = ''
        # find the start of the primer and output all following sequence
        try:
            match = re.search(fprimer, cseq)
            if keep_primers:
                forward_primer_seq = cseq[match.start():]
                fplen = match.end() - match.start()
            else:
                forward_primer_seq = cseq[match.end():]
        except AttributeError:
            forward_primer_seq = ''

        # if sequence can be amplified, search for reverse primer
        if forward_primer_seq != '':
            if rprimer is not None:
                # find the reverse primer match and use only up to it
                try:
                    match = re.search(rprimer, forward_primer_seq)
                    if keep_primers:
                        reverse_primer_seq = forward_primer_seq[:match.end()]
                    else:
                        reverse_primer_seq = forward_primer_seq[:match.start()]
                except AttributeError:
                    reverse_primer_seq = ''
            else:
                # reverse_primer_seq = forward_primer_seq[:length + fplen]
                reverse_primer_seq = forward_primer_seq[:]

            if reverse_primer_seq != '':
                # trim length if needed
                if length > 0:
                    if keep_primers:
                        reverse_primer_seq = reverse_primer_seq[:length + fplen]
                    else:
                        reverse_primer_seq = reverse_primer_seq[:length]
                printit = True

                if remove_ambig:
                    if 'N' in reverse_primer_seq:
                        printit = False

                if printit:
                    newseqs[ccseq] = reverse_primer_seq
    print('trimmed %d sequences, %d not found' % (len(newseqs), len(seqs) - len(newseqs)))
    return newseqs
